# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides a printer that transforms a list of Todo items to an iCalendar
file according to RFC 2445.
"""

import random
import string
from datetime import datetime, time

from topydo.lib.printers.PrettyPrinter import Printer


def _convert_priority(p_priority):
    """
    Converts todo.txt priority to an iCalendar priority (RFC 2445).

    Priority A gets priority 1, priority B gets priority 5 and priority C-F get
    priorities 6-9. This scheme makes sure that clients that use "high",
    "medium" and "low" show the correct priority.
    """
    result = 0

    prio_map = {
        'A': 1,
        'B': 5,
        'C': 6,
        'D': 7,
        'E': 8,
        'F': 9,
    }

    try:
        result = prio_map[p_priority]
    except KeyError:
        if p_priority:
            # todos with no priority have priority None, and result of this
            # function will be 0. For all other letters, return 9 (lowest
            # priority in RFC 2445).
            result = 9

    return result


class IcalPrinter(Printer):
    """
    A printer that converts a list of Todo items to a string in iCalendar
    format (RFC 2445).

    https://www.rfc-editor.org/rfc/rfc2445.txt
    """

    def __init__(self, p_todolist):
        super().__init__()
        self.todolist = p_todolist

        try:
            import icalendar
            self.icalendar = icalendar
        except ImportError:  # pragma: no cover
            self.icalendar = None

    def print_list(self, p_todos):
        result = ""

        if self.icalendar:
            cal = self.icalendar.Calendar()
            cal.add('prodid', '-//bramschoenmakers.nl//topydo//')
            cal.add('version', '2.0')

            for todo in p_todos:
                cal.add_component(self._convert_todo(todo))

            result = cal.to_ical().decode('utf-8')

        return result

    def _convert_todo(self, p_todo):
        """ Converts a Todo instance (Topydo) to an icalendar Todo instance. """

        def _get_uid(p_todo):
            """
            Gets a unique ID from a todo item, stored by the ical tag. If the
            tag is not present, a random value is assigned to it and returned.
            """

            def generate_uid(p_length=4):
                """
                Generates a random string of the given length, used as
                identifier.
                """
                return ''.join(
                    random.choice(string.ascii_letters + string.digits)
                    for i in range(p_length))

            uid = p_todo.tag_value('ical')
            if not uid:
                uid = generate_uid()
                p_todo.set_tag('ical', uid)
                self.todolist.dirty = True

            return uid

        result = self.icalendar.Todo()

        # this should be called first, it may set the ical: tag and therefore
        # change the source() output.
        result['uid'] = _get_uid(p_todo)

        result['summary'] = self.icalendar.vText(p_todo.text())
        result['description'] = self.icalendar.vText(p_todo.source())
        result.add('priority', _convert_priority(p_todo.priority()))

        start = p_todo.start_date()
        if start:
            result.add('dtstart', start)

        due = p_todo.due_date()
        if due:
            result.add('due', due)

        created = p_todo.creation_date()
        if created:
            result.add('created', created)

        completed = p_todo.completion_date()
        if completed:
            completed = datetime.combine(completed, time(0, 0))
            result.add('completed', completed)

        return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2015 Bram Schoenmakers <me@bramschoenmakers.nl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides a printer that transforms a list of Todo items to a graph in Dot
notation. Useful for displaying dependencies.
"""

from textwrap import wrap

from topydo.lib.printers.PrettyPrinter import Printer
from topydo.lib.ProgressColor import progress_color
from topydo.lib.Utils import humanize_date


class DotPrinter(Printer):
    """
    A printer that converts a list of Todo items to a string in Dot format.
    """

    def __init__(self, p_todolist):
        super(DotPrinter, self).__init__()
        self.todolist = p_todolist

    def print_list(self, p_todos):
        def node_label(p_todo):
            """
            Prints an HTML table for a node label with some todo details.
            """
            def escape_dot_label(p_string):
                """
                HTML like labels in Dot may not have raw ampersands, quotes or
                angle brackets. These should be properly replaced with the
                escaped character notation.
                """

                return p_string.replace('&', '&amp;').replace('"', '&quot;').replace(
                    '<', '&lt;').replace('>', '&gt;')

            node_result = '<<TABLE CELLBORDER="0" CELLSPACING="1" VALIGN="top">'

            def print_row(p_value1, p_value2):
                return '<TR><TD ALIGN="RIGHT">{}</TD><TD ALIGN="LEFT">{}</TD></TR>'.format(p_value1, p_value2)

            node_result += '<TR><TD><B>{}</B></TD><TD BALIGN="LEFT"><B>{}{}{}</B></TD></TR>'.format(
                self.todolist.number(p_todo),
                "<S>" if todo.is_completed() else "",
                "<BR />".join(map(escape_dot_label, wrap(p_todo.text(), 35))),
                "</S>" if todo.is_completed() else "",
            )

            priority = p_todo.priority()
            start_date = p_todo.start_date()
            due_date = p_todo.due_date()

            if priority or start_date or due_date:
                node_result += '<HR/>'

            if priority:
                node_result += print_row('Prio:', p_todo.priority())

            if start_date:
                node_result += print_row('Starts:', "{} ({})".format(
                    start_date.isoformat(),
                    humanize_date(start_date)
                ))

            if due_date:
                node_result += print_row('Due:', "{} ({})".format(
                    due_date.isoformat(),
                    humanize_date(due_date)
                ))

            node_result += '</TABLE>>'

            return node_result

        def foreground(p_background):
            """
            Chooses a suitable foreground color (black or white) given a
            background color.
            """

            (r, g, b) = p_background.as_rgb()
            brightness = (r * 299 + g * 587 + b * 114) / ( 255 * 1000 )

            return '#ffffff' if brightness < 0.5 else '#000000'

        node_name = lambda t: '_' + str(self.todolist.number(t))

        result = 'digraph topydo {\n'
        result += 'node [ shape="none" margin="0" fontsize="9" fontname="Helvetica" ]\n';

        # print todos
        for todo in p_todos:
            background_color = progress_color(todo)

            result += '  {} [label={} style=filled fillcolor="{}" fontcolor="{}"]\n'.format(
                node_name(todo),
                node_label(todo),
                background_color.as_html(),
                foreground(background_color),
            )

        # print edges
        for todo in p_todos:
            # only print the children that are actually in the list of todos
            children = set(p_todos) & set(self.todolist.children(todo,
                p_only_direct=True))

            for child in sorted(list(children), key=lambda t: t.text()):
                result += '  {} -> {}\n'.format(
                    node_name(todo),
                    node_name(child)
                )

        todos_without_dependencies = [todo for todo in p_todos if not self.todolist.children(todo) and not self.todolist.parents(todo)]
        for index in range(0, len(todos_without_dependencies) - 1):
            this_todo = todos_without_dependencies[index]
            next_todo = todos_without_dependencies[index + 1]
            result += '  {} -> {} [style="invis"]\n'.format(node_name(this_todo), node_name(next_todo))

        result += '}\n'
        return result



# Topydo - A todo.txt client written in Python.
# Copyright (C) 2014 - 2015 Bram Schoenmakers <bram@topydo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Provides functions to calculate the importance value of a task.

For those who are familiar with the Toodledo website, the importance value is a
combination of the priority and the todo's due date. Low priority tasks due
today may have a higher importance than high priority tasks in the distant
future.
"""

from datetime import date

from topydo.lib.Config import config

IMPORTANCE_VALUE = {'A': 3, 'B': 2, 'C': 1}


def is_due_next_monday(p_todo):
    """ Returns True when today is Friday (or the weekend) and the given task
    is due next Monday.
    """
    today = date.today()
    due = p_todo.due_date()

    return due and due.weekday() == 0 and today.weekday() >= 4 and \
        p_todo.days_till_due() <= 3


def importance(p_todo, p_ignore_weekend=config().ignore_weekends()):
    """
    Calculates the importance of the given task.
    Returns an importance of zero when the task has been completed.

    If p_ignore_weekend is True, the importance value of the due date will be
    calculated as if Friday is immediately followed by Monday. This in case of
    a todo list at the office and you don't work during the weekends (you
    don't, right?)
    """
    result = 2

    priority = p_todo.priority()
    result += IMPORTANCE_VALUE[priority] if priority in IMPORTANCE_VALUE else 0

    if p_todo.has_tag(config().tag_due()):
        days_left = p_todo.days_till_due()

        if days_left >= 7 and days_left < 14:
            result += 1
        elif days_left >= 2 and days_left < 7:
            result += 2
        elif days_left >= 1 and days_left < 2:
            result += 3
        elif days_left >= 0 and days_left < 1:
            result += 5
        elif days_left < 0:
            result += 6

    if p_ignore_weekend and is_due_next_monday(p_todo):
        result += 1

    if p_todo.has_tag(config().tag_star()):
        result += 1

    return result if not p_todo.is_completed() else 0


def average_importance(p_todo, p_ignore_weekend=config().ignore_weekends()):
    own_importance = importance(p_todo, p_ignore_weekend)

    average = 0
    parents = []

    try:
        sum_importance = own_importance
        parents = p_todo.parents()
        for parent in parents:
            sum_importance += importance(parent, p_ignore_weekend)

        average = float(sum_importance) / float(1 + len(parents))
    except AttributeError:
        pass

    return max(own_importance, average)
